#!/usr/bin/python3

'''
Adfinis SyGroup 2016
Author:       Patrick Winter
Version:      0.1
Description:  Check the values of JSON objects over http/https
Requirements: requests

Supported operators are:

    - eq --> equal
    - nt --> not equal
    # the following operators only work with numbers:
    - gt --> greater then
    - lt --> less then

'''

import re
import sys
import argparse
import requests

OPERATORS = {
    'eq': lambda a, b: a == b,
    'gt': lambda a, b: a > b,
    'lt': lambda a, b: a < b,
    'ne': lambda a, b: a != b,
}


def get_parsed_args():
    parser = argparse.ArgumentParser(
        'Queries JSON over http/https and checks a certain object')
    parser.add_argument('-U', '--url', action='store',
                        metavar='URL', type=str, required=True)
    parser.add_argument('-O', '--objectpath',
                        action='store', type=str, required=True)
    parser.add_argument('-v', '--value', action='store', type=str)
    parser.add_argument('-o', '--operator', action='store',
                        type=str, default='eq')
    parser.add_argument('-u', '--user', action='store', type=str)
    parser.add_argument('-p', '--password', action='store', type=str)
    parser.add_argument('-H', '--header', action='store', type=str, nargs='+')
    parser.add_argument('--no-ssl-verification', action='store_false')
    parser.set_defaults(ssl=False)
    return parser.parse_args()


def get_auth(args):
    "Take argparser-parsed args, and return an auth tuple (or None)"

    if args.user and args.password:
        return (args.user, args.password)

    return None


def do_request(url, auth, headers, verify_ssl):
    try:
        response = requests.get(
            url, auth=auth, verify=verify_ssl, headers=headers)

    except requests.exceptions.RequestException as e:
        print(e)
        sys.exit(3)

    valid_status_codes = [200, 301]
    if response.status_code not in valid_status_codes:
        print("Request returned none valid status code (%s)" %
              response.status_code)
        sys.exit(2)

    try:
        return response.json()
    except Exception as e:
        print('unable to parse json: %s' % str(e))
        sys.exit(3)


def is_int(value_str):
    return re.match(r'^\d+$', str(value_str))


def is_float(value_str):
    return re.match(r'^\d+\.\d+$', str(value_str))


def make_same_type(a, b):

    if is_int(a) and is_int(b):
        return (int(a), int(b))

    elif is_float(a) and is_float(b):
        return (float(a), float(b))

    elif is_int(a) and is_float(b) or is_float(a) and is_int(b):
        return (float(a), float(b))

    return (str(a), str(b))


def check_match(data, path, comparison_value, operator):
    for item in path.split('.'):
        if is_int(item) and isinstance(data, list):
            item = int(item)
        print(" .. evaluating path %s in dict %s" % (item, data))
        data = data[item]

    value1, value2 = make_same_type(data, comparison_value)

    return operator(value1, value2)


def parse_headers(headers):

    if headers:
        header_pairs = [
            re.split(': ', header, maxsplit=2)
            for header in headers
        ]

        return {
            key: value
            for key, value in header_pairs
        }

    return None


def main():
    args = get_parsed_args()

    response = do_request(
        url=args.url,
        auth=get_auth(args),
        headers=parse_headers(args.header),
        verify_ssl=args.no_ssl_verification
    )

    if args.operator not in OPERATORS:
        print("Unknown operator '%s'" % args.operator)
        sys.exit(3)

    operator = OPERATORS[args.operator]

    result = check_match(response, args.objectpath, args.value, operator)

    if result:
        sys.exit(0)
    sys.exit(2)


if __name__ == '__main__':
    main()
